/*

  yyysmartptr - a reference counting smart pointer implementation
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYSMARTPTR_HPP
#define __YYYSMARTPTR_HPP

namespace yyysmartptr {

  /*
    
    each object may only have ONE smartpointer tree point to it. example

    smartptr<foo> x (new foo);
    smartptr<foo> y (x); // ok.

    foo *i = new foo;
    smartptr<foo> a (i);
    smartptr<foo> b (i); // ILLEGAL AND VERY DANGEROUS !

    const smartptr<*> means you cannot change the pointer. but you can change the *.

    once an object has a smartpointer on it, it may never be deleted again!!!!!

  */

  template <class obj>
  class smartptr {
  public:
    smartptr (obj *const p_ = 0) : p(p_), count(0) { inc(); };
    smartptr (const smartptr &other) : p(other.p), count (other.count) { inc(); };

    smartptr & operator= (const smartptr &other)
    {
      if (this == &other) {
	// assignment to myself does nothing.
      }else{
	dec();
	p = other.p;
	count = other.count;
	inc();
      };
      return *this;
    };

    inline const bool isnull () const { return p==0; };

    void kill ()
    {
      dec();
      p = 0;
      count = 0;
      inc();
    };

    ~smartptr () { dec(); };

    inline obj *const operator-> () const { return p; };
    inline obj & operator* () const { return *p; };

  private:
    void dec()
    {
      if (count != 0) {
	if ((*count) != 0) {
	  (*count)--;
	  if ((*count) == 0) {
	    delete p;
	    delete count;
	    count = 0;
	  };
	};
      };
    };
    void inc()
    { 
      if (count != 0) {
	(*count)++;
      }else{
	count = new unsigned int(1);
      };
    };

    obj *p;
    unsigned int *count;
  };

};

#endif
