#include "yyypostman.hpp"
#include "binarydefs.hpp"
#include "memchk.hpp"
#include "fatal.hpp"

#include <string>
#include <iostream>
#include <vector>
#include <unistd.h>

using namespace yyypostman;

// define the class "bb". it is a model of the binbuffer concept
class bb {
public:
  typedef int32 index;
  inline const uint8 & operator [] (const index x) const { return rep[x]; };
  inline uint8 & operator [] (const index x) { return rep[x]; };
  inline const index size() const { return rep.size(); };
  inline void resize(const index s) { rep.resize(s); };
  inline void erase (const index first, const index last) { rep.erase(&rep[first], &rep[last]); };
private:
  vector<uint8>rep;
};

// define the class net. it is a model of the binstream concept
// instead of actually sending data to the internet, it just puts
// all data to be sent into the receive buffer.
// it also prints all data onto screen.
// so "net" is a debugging loopback device.
class net {
public:
  typedef bb binbuffer;

  binbuffer receivebuffer, sendbuffer;
  SigC::Signal1<void,net&> receive;
  void send ()
  {
    // first we put the contents of the sendbuffer onto screen
    cout << "Sending: [ ";
    for (bb::index i=0;i<sendbuffer.size();++i) cout << int(sendbuffer[i]) << ' ';
    cout << "]\n";
    // then we append the sendbuffer to the receivebuffer and erase the sendbuffer.
    bb::index inspos=receivebuffer.size();
    receivebuffer.resize(inspos+sendbuffer.size());
    for (bb::index i=0;i<sendbuffer.size();++i) receivebuffer[inspos+i]=sendbuffer[i];
    sendbuffer.erase(0,sendbuffer.size());
    receive.emit(*this); // we must announce that there is new data.
  };
};


// define the class upperstring. it is a model of the message concept
class upperstring {
  friend void appendtobinbuffer (const upperstring &u, bb &x);
  friend const bb::index readfrombinbuffer (upperstring &u, const bb &x, const bb::index startpos);
public:
  upperstring () {};
  upperstring (const string &x)
  {
    if (!isvalid(x)) fatal("upperstring: constructor: string invalid");
    str=x;
  };
  const string & get() const { return str; };
  const bool isvalid (const string &x) const
  {
    for (string::const_iterator i=x.begin();i!=x.end();++i) if (!isvalidchar(*i)) return false;
    return true;
  };
  const bool isvalidchar (const char x) const
  {
    return (x>=32) && (x<=126);
  };
private:
  string str;
};

void appendtobinbuffer (const upperstring &u, bb &x)
{
  bb::index inspos=x.size();
  x.resize(inspos+u.str.size()+1);
  bb::index i;
  for (i=0;i<bb::index(u.str.size());++i) x[inspos+i]=u.str[i];
  x[inspos+i]=234; // << end of string mark
};

const bb::index readfrombinbuffer (upperstring &u, const bb &x, const bb::index startpos)
{
  u.str.erase();
  bb::index p=startpos;
  while (p<x.size()) {
    if (x[p]==234) {
      return p-startpos+1;
    }else if (u.isvalidchar(x[p])) {
      u.str+=x[p++];
    }else{
      fatal("upperstring: illegal chars received");
    };
  };
  return -1;
};


// define the class fourbyte. it is a model of the message concept
class fourbyte {
  friend void appendtobinbuffer (const fourbyte &f, bb &x);
  friend const bb::index readfrombinbuffer (fourbyte &f, const bb &x, const bb::index startpos);
public:
  fourbyte () {};
  fourbyte (const uint8 a_, const uint8 b_, const uint8 c_, const uint8 d_)
    : a(a_), b(b_), c(c_), d(d_) {};
  void dump() const { cout << "[A="<<int(a) << " B="<<int(b) << " C="<<int(c) << " D="<<int(d) << "]"; };
  inline const uint8 get_a() const { return a; };
  inline const uint8 get_b() const { return b; };
  inline const uint8 get_c() const { return c; };
  inline const uint8 get_d() const { return d; };
private:
  uint8 a,b,c,d;
};

void appendtobinbuffer (const fourbyte &f, bb &x)
{
  bb::index inspos=x.size();
  x.resize(inspos+4);
  x[inspos]=f.a;
  x[inspos+1]=f.b;
  x[inspos+2]=f.c;
  x[inspos+3]=f.d;
};

const bb::index readfrombinbuffer (fourbyte &f, const bb &x, const bb::index startpos)
{
  if (x.size()-startpos<4) return -1;
  f.a=x[startpos];
  f.b=x[startpos+1];
  f.c=x[startpos+2];
  f.d=x[startpos+3];
  return 4;
};


// define the class address. it is a model of the messagekind concept
class address {
  friend void appendtobinbuffer (const address &a, bb &b);
  friend const bb::index readfrombinbuffer(address &a, const bb &b, const bb::index pos);
public:
  address() {};
  address(const uint8 v_) : v(v_) {};
  const bool operator < (const address &o) const { return v < o.v; };
private:
  uint8 v;
};

void appendtobinbuffer (const address &a, bb &b)
{
  bb::index pos=b.size();
  b.resize(pos+1);
  b[pos]=a.v;
};

const bb::index readfrombinbuffer(address &a, const bb &b, const bb::index pos)
{
  if (b.size()-pos<1) return -1;
  a.v=b[pos];
  return 1;
};


void appendtobinbuffer (const int a, bb &b)
{
  uint32 r=a;
  fourbyte f(r&255,(r>>8)&255,(r>>16)&255,(r>>24)&255);
  appendtobinbuffer(f,b);
};

const bb::index readfrombinbuffer(int &a, const bb &b, const bb::index pos)
{
  fourbyte f;
  bb::index i=readfrombinbuffer(f,b,pos);
  uint32 r;
  r=(uint32(f.get_a())) | (uint32(f.get_b())<<8) | (uint32(f.get_c())<<16) | (uint32(f.get_d())<<24);
  a=r;
  return i;
};


class testbed : public SigC::Object {
public:
  testbed () : post(&mynet) {};
  void myrecv (const upperstring &e)
  {
    cout << "myrecv [" << e.get() << "]. resending on 200 !\n";
    post.sender1<const upperstring&> (200) (e);
  };
  
  void myrecvcool (const upperstring &e)
  {
    cout << "myrecvcool [" << e.get() << "]\n";
  };
  
  void myrecvfb (const fourbyte &e)
  {
    cout << "myrecvfb ";
    e.dump();
    cout << endl;
  };

  void mydualrecv (const fourbyte &fb, const upperstring &us)
  {
    cout << "mydualrecv got the fourbyte ";
    fb.dump();
    cout << " and the string [" << us.get() << "]\n";
  };
  
  void myzeroparmrecv ()
  {
    cout << "i got a zero parameter msg\n";
  };

  void intreceiver (int i)
  {
    cout << "YES I RECEIVED A NORMAL INTEGER !!!!" << i << "\n";
  };

  void receivingerror (postman<address,net> &)
  {
    cout << "the receiving stream is closed due to error.\n";
  };

  void dotest()
  {
    post.corruptreceived.connect(SigC::slot(*this,&testbed::receivingerror));

    post.receiver1_open<const upperstring &> (100) .connect (SigC::slot(*this,&testbed::myrecv));
    post.receiver1_open<const upperstring &> (200) .connect (SigC::slot(*this,&testbed::myrecvcool));
    post.receiver1_open<const fourbyte &> (150) .connect (SigC::slot(*this,&testbed::myrecvfb));
    post.receiver2_open<const fourbyte&,const upperstring&> (50); // this time we first open, then connect.
    // just to show that you can always get your signal object after opening, too.
    post.receiver2<const fourbyte&,const upperstring&> (50) .connect (SigC::slot(*this,&testbed::mydualrecv));
    post.receiver0_open (222) .connect (SigC::slot(*this,&testbed::myzeroparmrecv));
    post.receiver1_open<int> (250) .connect (SigC::slot(*this,&testbed::intreceiver));

    post.sender1_open<const upperstring&> (100);
    post.sender1_open<const upperstring&> (200);
    post.sender1_open<const fourbyte&> (150);
    post.sender2_open<const fourbyte&,const upperstring&> (50);
    post.sender0_open (222) (); // <-- you get the slot object also already when opening
    post.sender1_open<int> (250);

    post.sender1<const upperstring&> (100) (upperstring("ABCDEFG"));
    post.sender1<const upperstring&> (200) (upperstring("Ain't that cool ?"));
    post.sender1<const fourbyte&> (150) (fourbyte(4,5,6,7));
    post.sender1<const upperstring&> (100) (upperstring("Sure it is !"));
    post.sender2<const fourbyte&,const upperstring&> (50) (fourbyte(10,20,30,40), upperstring("suxxerstring"));
    post.sender0 (222) ();
    post.sender1<int> (250) (44000);

    post.senderx_close (222);
    post.senderx_close (50);
    post.senderx_close (100);
    post.senderx_close (200);
    post.senderx_close (150);

    post.receiverx_close (250);
    post.receiverx_close (222);
    post.receiverx_close (50);
    post.receiverx_close (100);
    post.receiverx_close (200);
    post.receiverx_close (150);
  };
private:
  net mynet;
  postman<address,net> post; // postman must die FIRST !!!!
};


int main()
{
  memchk::resetsituation();
  {

    try {
      testbed b;
      b.dotest();
    } catch (const char *const x) {
      fatal(x);
    };

  }
  if (!memchk::everythingzero()) memchk::printstatistics(cerr);

  return 0;
};
